module ("L_HomeCare", package.seeall)

------------------------------------------------------------------------------------------------------------------------
-- Constants
------------------------------------------------------------------------------------------------------------------------
local TIME_BETWEEN_NOTIFICATIONS = 1800 -- seconds / 30 minutes
local MIN_TIME_BETWEEN_NOTIF = 300 -- seconds / 5 minutes
local CHECK_INTERVAL = "1m"

local SID = {
	["HOME_CARE"] = "urn:micasaverde-com:serviceId:HomeCare1",
	["SES"]       = "urn:micasaverde-com:serviceId:SecuritySensor1"
}

local SYS_MESSAGE_TYPES = {
	BUSY  = 1,
	ERROR = 2,
	CLEAR = 4
}

------------------------------------------------------------------------------------------------------------------------
-- Flags
------------------------------------------------------------------------------------------------------------------------
local F_DEBUG_MODE = true

------------------------------------------------------------------------------------------------------------------------
-- Global variables
------------------------------------------------------------------------------------------------------------------------
local lug_device
local g_taskHandle = -1    -- Handle for the luup.task function
local g_lastSysMessage = 0 -- Time of the last system message
local notifyNr = 1
local g_deviceData = {
	-- period:        The inactivity period (in minutes)
	-- intervals:     Stores the time intervals when the device will send notification if tripped.
	--                It's an array where each item has this structure: { start, end_ }
	-- enabled:       If 'true', the device will send notifications.
}
local g_users = {} -- The list of users on the unit, indexed by username

------------------------------------------------------------------------------------------------------------------------
-- Misc functions
------------------------------------------------------------------------------------------------------------------------
local log = luup.log
local function debug() end


local function ShowSysMessage (message, mode, permanent)
	mode = mode or SYS_MESSAGE_TYPES.BUSY
	permanent = permanent or false
	--log( "(HomeCare::ShowSysMessage-"..(getinfo(2).name or "N/A")..") mode: ".. mode ..", permanent: ".. tostring( permanent ) ..", message: ".. message )
	luup.task( message, mode, "HomeCare", g_taskHandle )
	g_lastSysMessage = os.time()

	if not permanent then
		-- Clear the previous system message, since it's transient.
		luup.call_delay( "ClearSysMessage", 30, g_lastSysMessage )
	elseif mode == SYS_MESSAGE_TYPES.ERROR then
		-- Critical error.
		luup.set_failure( true, lug_device )
	end
end


function ClearSysMessage (messageTime)
	-- 'messageTime' is nil if the function is called by the user.
	if (messageTime == g_lastSysMessage) or (messageTime == nil) then
		luup.task( "Clearing...", SYS_MESSAGE_TYPES.CLEAR, "HomeCare", g_taskHandle )
	end
end


-- Splits a string by the given delimiter, which by default is whitespace.
local function SplitString (str, delimiter)
	delimiter = delimiter or "%s+"
	local result = {}

	local from = 1
	local delimFrom, delimTo = str:find( delimiter, from )
	while delimFrom do
		table.insert( result, str:sub( from, delimFrom-1 ) )
		from = delimTo + 1
		delimFrom, delimTo = str:find( delimiter, from )
	end
	table.insert( result, str:sub( from ) )

	return result
end


local function UrlEncode (str)
	str = str:gsub( "\n", "\r\n" )
	str = str:gsub("([^%w])", function (c)
	                              return string.format( "%%%02X", string.byte(c) )
	                          end)
	return str
end


local function PadLeft (s, length, c)
	s = tostring(s)
	length = length or 2
	c = c or "0"
	return c:rep( length - #s ) .. s
end

local function checkVersion()
	if( luup.version_branch ~= 1 or luup.version_major ~= 7 ) then
		return false
	else
		return true
	end
end

-- Get the Vera IP address based on the alarm panel's IP address.
-- If the camera is not in Vera's LAN, assume it's in the Vera's WAN.
-- Currently works only if the netmask is 255.255.255.0.
-- If the IP address couldn't be detected, return empty string.
local function FindAddress()
	local gatewayAddress = ""
	local stdout = io.popen("GetNetworkState.sh ip_wan")
	local wanIp = stdout:read("*a")
	stdout:close()
	if wanIp then
		gatewayAddress = wanIp
	end
	debug("(HomeCare)::(FindAddress) : Return address: ".. gatewayAddress)
	return gatewayAddress
end
------------------------------------------------------------------------------------------------------------------------
-- Plugin specific functions
------------------------------------------------------------------------------------------------------------------------
function SendAlert(devNum)
	debug( "(HomeCare)::(SendAlert) : Device: ".. devNum .." sending notification" )

	----------------------------------------------------------------------------
	-- Get the users to notify.
	----------------------------------------------------------------------------
	local userList = luup.variable_get( SID.HOME_CARE, "UserList", lug_device ) or "0"
	if userList == "0" or userList == "" then
		log( "(HomeCare)::(SendAlert) : No users to notify, return" )
		return
	end
	-- Get the user ids.
	local userIds = ""
	for userName in userList:gmatch("[^,]+") do
		if g_users[userName] then
			userIds = userIds ..",".. g_users[userName]
		end
	end
	userIds = userIds:sub( 2, -1 )
	local unitIp = FindAddress()
	if unitIp == "" then
		debug("(HomeCare)::(SendAlert) : ERROR :  Cannot detect unit IP, cannot set alerts!")
	else
		local url = "http://"..unitIp..":3480/data_request?id=add_alert&device="..devNum.."&type=3&source=3&severity=0&code="..UrlEncode("HomeCare").. "&description="..UrlEncode( luup.devices[devNum].description .." (#".. devNum ..") hasn't been tripped").."&users="..userIds
		local status, content = luup.inet.wget(url,15)
		if not content then
			debug("(HomeCare)::(SendAlert) : ERROR :  Alert could NOT be set!")
		else
			debug("(HomeCare)::(SendAlert) : OK :  Alert has been set!")
		end
	end
end

function SendNotification (devNum)
	debug( "(HomeCare::SendNotification) Device: ".. devNum .." sending notification" )

	----------------------------------------------------------------------------
	-- Get the users to notify.
	----------------------------------------------------------------------------
	local userList = luup.variable_get( SID.HOME_CARE, "UserList", lug_device ) or "0"
	if userList == "0" or userList == "" then
		log( "(HomeCare::SendNotification) No users to notify, return" )
		return
	end
	-- Get the user ids.
	local userIds = ""
	for userName in userList:gmatch("[^,]+") do
		if g_users[userName] then
			userIds = userIds ..",".. g_users[userName]
		end
	end
	userIds = userIds:sub( 2, -1 )

	----------------------------------------------------------------------------
	-- Compose the notification URL.
	----------------------------------------------------------------------------
	devNum = tonumber( devNum, 10 )
	local url = "https://".. luup.event_server .."/alert?" ..
	            "PK_AccessPoint=".. string.format("%d", luup.pk_accesspoint) ..
	            "&HW_Key=".. luup.hw_key ..
	            "&DeviceID=".. lug_device ..
	            "&LocalDate=".. os.date("%Y-%m-%d%%20%H:%M:%S") ..
	            "&LocalTimestamp=".. os.time() ..
	            "&AlertType=3&SourceType=3"..
	            "&Description=".. UrlEncode( luup.devices[devNum].description .." (#".. devNum ..") hasn't been tripped" ) ..
	            "&Users=".. UrlEncode( userIds )
	debug( "(HomeCare::SendNotification) Notification URL: ".. url )

	-- Send the notification.
	local status = luup.inet.wget( url, 10 )
	if status ~= 0 then
		log( "(HomeCare::SendNotification) ERROR: Failed to send notificaton" )
	else
		log( "(HomeCare::SendNotification) SUCCESS: Notification sent" )
		luup.variable_set( SID.HOME_CARE, "LastNotification", os.time(), devNum )
	end
end


function CheckLastTrip()
	local timeBetweenNotif = luup.variable_get( SID.HOME_CARE, "TimeBetweenNotif", lul_device ) or ""
	if timeBetweenNotif == "" then
		luup.variable_set( SID.HOME_CARE, "TimeBetweenNotif", TIME_BETWEEN_NOTIFICATIONS, lul_device)
	else
		timeBetweenNotif = tonumber( timeBetweenNotif, 10 )
		if timeBetweenNotif >= MIN_TIME_BETWEEN_NOTIF then
			TIME_BETWEEN_NOTIFICATIONS = timeBetweenNotif
		else
			ShowSysMessage( "Time between notification can't be smaller than ".. (MIN_TIME_BETWEEN_NOTIF / 60) .." minutes. Using default", SYS_MESSAGE_TYPES.BUSY )
			debug( "(HomeCare::CheckLastTrip) Time between notification can't be smaller than ".. (MIN_TIME_BETWEEN_NOTIF / 60) .." minutes. Using default" )
		end
	end
	for devNum in pairs( g_deviceData ) do

		--------------------------------------------------------------------------------------------
		-- Check if notifications are enabled for this device.
		--------------------------------------------------------------------------------------------
		if g_deviceData[devNum].enabled then
			debug( "(HomeCare::CheckLastTrip) Device #".. devNum )
			----------------------------------------------------------------------------------------
			-- Check if enough time has passed since the last notification.
			----------------------------------------------------------------------------------------
			local lastNotification = luup.variable_get( SID.HOME_CARE, "LastNotification", devNum ) or 0
			if os.difftime( os.time(), lastNotification ) >= TIME_BETWEEN_NOTIFICATIONS then
				debug( "(HomeCare::CheckLastTrip) Enough time has passed since the last notification. Continue" )

				------------------------------------------------------------------------------------
				-- Check if the current time is in this device's activity interval.
				------------------------------------------------------------------------------------
				local intIdx = {} -- The active interval index
				local intervalsIndex = -100
				local now = os.date("*t")
				now = now.hour * 100 + now.min
				for idx, interval in pairs( g_deviceData[devNum].intervals ) do
					local start = {} 
					local end_ = {} 
					local offset = 1
					local len = 0
					local count = 1
					for k, v in pairs(interval) do
						len = len + 1
					end
					while offset <= len do
						start[offset] = tonumber(tostring(PadLeft(interval[offset]) .. tostring(PadLeft(interval[offset + 1]))))
						end_[offset] = tonumber(tostring(PadLeft(interval[offset + 2]) .. tostring(PadLeft(interval[offset + 3]))))
						offset =  offset + 4
					end
					for index, value in pairs(start) do
						if value < end_[index] then
							-- The time interval is in the same day.
							if (now >= value) and (now <= end_[index]) then
								intIdx[count] = index
								intervalsIndex = tonumber(idx)
								count = count + 1
							end
						else
							-- The time interval spans on two days.
							if (now >= value) or (now <= end_[index]) then
								intIdx[count] = index
								intervalsIndex = tonumber(idx)
								count = count + 1
							end
						end
					end
				end
				for key, value in pairs(intIdx) do
					if value then
						debug( "(HomeCare::CheckLastTrip) We are in the notifications interval. Continue" )

						--------------------------------------------------------------------------------
						-- Check how many minutes have passed since the active interval began, to reset the LastTrip variable,
						-- because it is assumed that no motion occured during the inactivity period.
						--------------------------------------------------------------------------------
						local checkMin = tonumber( CHECK_INTERVAL:match("^%d+"), 10 )
						local startMin = g_deviceData[devNum].intervals[intervalsIndex][value + 1] --% 100
						now = os.date("*t")
						-- If the current minute is in a different hour, add 60 to get a positive difference.
						if now.min < startMin then
							now.min = now.min + 60
						end
						local diff = now.min - startMin -- How many minutes have passed since the activity interval began
						local lastTripV = luup.variable_get( SID.SES, "LastTrip", devNum ) or "-1"
						if diff < checkMin or lastTripV == "-1" then
							debug( "(HomeCare::CheckLastTrip) Set LastTrip to the start of activity interval" )
							luup.variable_set( SID.SES, "LastTrip", os.time() - (diff * 60), devNum )
						end

						--------------------------------------------------------------------------------
						-- Check if enough time has passed since the sensor was last tripped.
						--------------------------------------------------------------------------------
						local lastTrip = luup.variable_get( SID.SES, "LastTrip", devNum ) or os.time()
						if os.difftime( os.time(), lastTrip ) >= g_deviceData[devNum].period[intervalsIndex] * 60 then
							----------------------------------------------------------------------------
							-- All the conditions are met. Send notification.
							----------------------------------------------------------------------------
							debug( "(HomeCare::CheckLastTrip) Enough time has passed since the sensor was last tripped. Send notification" )
							if checkVersion() then
								SendAlert(devNum)
							else
								SendNotification( devNum )
							end
						end
					end -- if value
				end -- for intIdx
			end -- if difftime >= TIME_BETWEEN_NOTIFICATIONS
		end -- if g_deviceData[devNum].enabled
	end -- for

	luup.call_timer( "CheckLastTrip", 1, CHECK_INTERVAL, "" )
end


local function UpdateDeviceData()
	local deviceData = ""
	local items = {}
	for devNum, info in pairs( g_deviceData ) do
		-- dDEVICE_ID,iINACTIVITY_PERIOD,sPERIOD1_START,ePERIOD1_END,...,sPERIODn_START,ePERIODn_END,nNOTIFICATIONS_ENABLED
		local item = "d".. devNum
		for indexp, periodI in pairs(info.period) do
			item = item ..",i".. PadLeft(periodI, 2)
		end
		--item = item ..",i".. info.period
		for indexa, interval in ipairs( info.intervals ) do
			local intervalLength = 0
			for k in pairs(interval) do
				intervalLength = intervalLength + 1
			end
			local val = 1;
			while val <= intervalLength do
				item = item ..",sh".. PadLeft( interval[val], 2 ) .. ",sm".. PadLeft( interval[val+1], 2 ) ..",eh".. PadLeft( interval[val+2], 2 ) ..",em".. PadLeft( interval[val+3], 2 )
				val = val + 4
			end
		end
		--local interval = info.intervals
		--local val = 0;
		item = item ..",n".. (info.enabled and "1" or "0")
		table.insert( items, item )
		deviceData = table.concat( items, ';' )
	end
	debug( "(HomeCare::UpdateDeviceData) New device data: ".. deviceData )
	luup.variable_set( SID.HOME_CARE, "DeviceData", deviceData, lug_device )
end


local function GetDeviceData (DeviceData)
	-- Parse the DeviceData variable.
	local deviceData = SplitString( DeviceData, ';' )
	for _, info in pairs( deviceData ) do
		local devNum = tonumber( info:match("d(%d+)"), 10 )
		-- Check if the device still exists and it's the same.
		if luup.devices[devNum] and (luup.devices[devNum].category_num == 4) then
			g_deviceData[devNum] = {}
			g_deviceData[devNum].period = {}
			--g_deviceData[devNum].period = tonumber( info:match("i(%d+)"), 10 )
			for v in string.gmatch(info,"i(%d+),") do
				table.insert(g_deviceData[devNum].period, tonumber(v, 10))
				notifyNr = notifyNr + 1
			end
			-- Get the intervals.
			g_deviceData[devNum].intervals = {}
			local array = {}
			for start_h, start_m, end_h, end_m in info:gmatch( "sh(%d+),sm(%d+),eh(%d+),em(%d+)" ) do
				start_h = tonumber( start_h, 10 )
				start_m = tonumber( start_m, 10 )
				end_h = tonumber( end_h, 10 )
				end_m = tonumber( end_m, 10 )
				table.insert(array, start_h)
				table.insert(array, start_m)
				table.insert(array, end_h)
				table.insert(array, end_m )
				--table.insert( g_deviceData[devNum].intervals, { start_h, start_m, end_h, end_m } )
			end
			table.insert( g_deviceData[devNum].intervals, array )
			g_deviceData[devNum].enabled = (info:match("n(%d)") == "1")
		else
			debug( "(HomeCare::Init) Device #".. devNum .." doesn't exist anymore, or it has changed" )
		end
	end
end


local function GetUsers()
	-- username=1|user_id
	local numUsers = 0
	local userName, userId
	for line in io.lines( "/etc/cmh/users.conf" ) do
		userName, userId = line:match( "^([^=]+)=%d|(%d+)" )
		if userName and userId then
			g_users[userName] = userId
			numUsers = numUsers + 1
		end
	end
	return numUsers
end


local function GetDebugMode()
	local debugMode = luup.variable_get( SID.HOME_CARE, "DebugMode", lug_device ) or ""
	if debugMode == "" then
		debugMode = F_DEBUG_MODE and "1" or "0"
		luup.variable_set( SID.HOME_CARE, "DebugMode", debugMode, lug_device )
	else
		DEBUG_MODE = ( debugMode == "1" )
	end
end


function DeviceDataWatchCallback (device, serviceId, variable, oldValue, newValue)
	GetDeviceData( newValue )
end


------------------------------------------------------------------------------------------------------------------------
-- Main
------------------------------------------------------------------------------------------------------------------------

function Init (lul_device)

	lug_device = lul_device

	GetDebugMode()
	if DEBUG_MODE then
		debug = log
	end

	-- Get the list of users from users.conf.
	local numUsers = GetUsers()
	if( numUsers == 0 ) then
		log( "(HomeCare::Init) No users in users.conf, return" )
		return
	end

	------------------------------------------------------------------------------------------------
	-- Create and initialize the UserList state variable.
	------------------------------------------------------------------------------------------------
	local userList = luup.variable_get( SID.HOME_CARE, "UserList", lug_device )
	if not userList then
		debug( "(HomeCare::Init) UserList is nil, initialize it to 0" )
		luup.variable_set( SID.HOME_CARE, "UserList", "0", lug_device )
	end

	------------------------------------------------------------------------------------------------
	-- Read the TimeBetweenNotif variable. If it doesn't exist, create it and assign it a default value.
	------------------------------------------------------------------------------------------------
	local timeBetweenNotif = luup.variable_get( SID.HOME_CARE, "TimeBetweenNotif", lul_device ) or ""
	if timeBetweenNotif == "" then
		luup.variable_set( SID.HOME_CARE, "TimeBetweenNotif", TIME_BETWEEN_NOTIFICATIONS, lul_device)
	else
		timeBetweenNotif = tonumber( timeBetweenNotif, 10 )
		if timeBetweenNotif >= MIN_TIME_BETWEEN_NOTIF then
			TIME_BETWEEN_NOTIFICATIONS = timeBetweenNotif
		else
			ShowSysMessage( "Time between notification can't be smaller than ".. (MIN_TIME_BETWEEN_NOTIF / 60) .." minutes. Using default", SYS_MESSAGE_TYPES.BUSY )
		end
	end

	------------------------------------------------------------------------------------------------
	-- Create or update the DeviceData state variable.
	-- This variable is a list separated by semicolons( ; ).
	-- Each item in the list is also a list separated by commas( , ).
	-- The items in the sublist are:
	--  dDEVICE_ID,iINACTIVITY_PERIOD,sPERIOD1_START,ePERIOD1_END,...,sPERIODn_START,ePERIODn_END,nNOTIFICATIONS_ENABLED
	-- The period start and end are expressed as military time, e.g. 06:30 = 630 (or 0630), 17:25 = 1725
	-- Example:d8,i3,s600,e830,s1125,e1745,n1;d12,i7,s930,e-1300,s1905,e2305,n0
	------------------------------------------------------------------------------------------------
	local deviceData = luup.variable_get( SID.HOME_CARE, "DeviceData", lul_device ) or ""
	debug( "(HomeCare::Init) Current device data: ".. deviceData )
	if deviceData == "" then
		-- Get the list of security sensors.
		for devNum, devAttr in pairs( luup.devices ) do
			if devAttr.category_num == 4 then
				-- The device is a security sensor.
				debug( "(HomeCare::Init) Adding device #".. devNum .."-".. devAttr.description )
				g_deviceData[devNum] = {}
				g_deviceData[devNum].period = {0}
				notifyNr = notifyNr + 1
				g_deviceData[devNum].intervals = { { 0, 59, 0, 59 } }
				g_deviceData[devNum].enabled = false
			end
		end
	else
		-- Get the device data from the DeviceData variable.
		GetDeviceData( deviceData )
		-- Update our list of security sensors.
		for devNum, devAttr in pairs( luup.devices ) do
			if devAttr.category_num == 4 and not g_deviceData[devNum] then
				-- This security sensor isn't in the list.
				debug( "(HomeCare::Init) Adding device #".. devNum .."-".. devAttr.description )
				g_deviceData[devNum] = {}
				g_deviceData[devNum].period = {0}
				g_deviceData[devNum].intervals = { { 0, 59, 0, 59 } }
				g_deviceData[devNum].enabled = false
			end
		end
	end

	-- Update the DeviceData variable.
	UpdateDeviceData()

	-- Add watch for the DeviceData variable, to update g_deviceData without restarting LuaUPnP.
	luup.variable_watch( "DeviceDataWatchCallback", SID.HOME_CARE, "DeviceData", lul_device )

	-- Check regularly if any sensor was tripped, starting from now.
	CheckLastTrip()
	log( "(HomeCare::Init) Startup successful" )
	return true
end
